#----------------------------------------------------------------------------#
# Copyright 2010, Regular Polygon

# Permission to use this software without fee is hereby granted.
# Distribution or publication of this software, or substantial portions of
# the software, is subject to the expressed written consent of the author.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#----------------------------------------------------------------------------#
# Name        :   Superellipsoid
# Filename    :   rp_superellipsoid_v10.rb
# Version     :   1.0
# SU Version  :   6.0+
# Type        :   Dialog Box
# Description :   Creates a superellipsoid component. 
# Menu Item   :   Draw -> Superellipsoid
# Context Menu:   N/A
# Usage       :   Fill in the dialog box.
# Date        :   7/1/2010
# Author      :   Doug Herrmann
# Email       :   regular.polygon@gmail.com
# Website     :   http://regularpolygon.blogspot.com/
#----------------------------------------------------------------------------#


require 'sketchup.rb'


module RP

#----------------------------------------------------------------------------#
class Superellipsoid

	# Default dialog box parameters
	@@x = (Sketchup.read_default 'RP_superellipsoid', 'x', '24').to_f.inch      # x axis length
	@@y = (Sketchup.read_default 'RP_superellipsoid', 'y', '24').to_f.inch      # y axis width
	@@z = (Sketchup.read_default 'RP_superellipsoid', 'z', '24').to_f.inch      # z axis width
	@@hs = (Sketchup.read_default 'RP_superellipsoid', 'hs', '50').to_i         # horizontal squareness
	@@vs = (Sketchup.read_default 'RP_superellipsoid', 'vs', '50').to_i         # vertical squareness
	@@rings = (Sketchup.read_default 'RP_superellipsoid', 'rings', '17').to_i   # number of latitude lines
	@@edges = (Sketchup.read_default 'RP_superellipsoid', 'edges', '24').to_i   # number of longitude lines (edges per latitude line)
	

	def initialize(x, y, z, hs, vs, rings, edges)
		@a, @b, @c = (x / 2.0), (y / 2.0), (z / 2.0)
		@h_exp, @v_exp = sqr_to_exp(hs), sqr_to_exp(vs)
		@rings, @edges = rings, edges

		# make geomentry creation undoable
		Sketchup.active_model.start_operation 'Definition'  
			 create_definition
		Sketchup.active_model.commit_operation 
	end


	# This method handles the Draw -> Superellipsoid menu item.
	def self.dialog

		# Show dialog to get user input for superellipsoid paramters
		prompts = [
			'Length x-axis: ', 'Width y-axis:  ', 'Height z-axis: ', 
			'Horz Squareness: ', 'Vert Squareness: ', 
			'Latitude Lines: ', 'Longitude Lines: '
		]
		defaults = [@@x, @@y, @@z, @@hs, @@vs, @@rings, @@edges]
		results = nil

		# Keep prompting user until input is valid.
		begin
			results = UI.inputbox prompts, defaults, 'Superellipsoid Parameters'
			return unless results  # user cancelled the operation
			x, y, z, hs, vs, rings, edges = results
			raise "Invalid length: #{ x }"  if ( x <= 0 )
			raise "Invalid width: #{ y }"   if ( y <= 0 )
			raise "Invalid height: #{ z }"  if ( z <= 0 )
			raise "Enter a horizontal squareness value between -99 and +99."  if ( (hs > 99) || (hs < -99) )
			raise "Enter a vertical squareness value between -99 and +99."  if ( (vs > 99) || (vs < -99) )
			raise "Enter at least 1 line of latitude."  if ( rings < 1 )
			raise "Enter at least 4 lines of longitude."  if ( edges < 4 )
		rescue
			UI.messagebox $!.message
			retry
		end
		
		# Create the superellipsoid component and place it in the model.
		self.new(*results).place_component

		# Remember parameters
		@@x, @@y, @@z, @@hs, @@vs, @@rings, @@edges = results

		# Remember parameters between sessions
		Sketchup.write_default 'RP_superellipsoid', 'x', @@x.inspect
		Sketchup.write_default 'RP_superellipsoid', 'y', @@y.inspect
		Sketchup.write_default 'RP_superellipsoid', 'z', @@z.inspect
		Sketchup.write_default 'RP_superellipsoid', 'hs', @@hs.to_s
		Sketchup.write_default 'RP_superellipsoid', 'vs', @@vs.to_s
		Sketchup.write_default 'RP_superellipsoid', 'rings', @@rings.to_s
		Sketchup.write_default 'RP_superellipsoid', 'edges', @@edges.to_s
	end


	# Place an instance in the model using the component placement tool.
	def place_component
		Sketchup.active_model.place_component @definition 
	end

	
	private  #-----------------------------------------------------------------#


	# map squareness scale value the exponent value
	def sqr_to_exp(s)
		root2 = 1 / Math.sqrt( 2 )
		if ( s >= 0 )
			exp = -Math.log(2) / Math.log( root2 + s * (1 - root2) / 100 )
		else
			exp = -Math.log(2) / Math.log( root2 - s * (0.5 - root2) / 50 )
		end
	end


	# Create a superellipsoid component definition
	def create_definition
		@definition = Sketchup.active_model.definitions.add 'Superellipsoid'
		@definition.entities.fill_from_mesh mesh
		@definition.insertion_point = Geom::Point3d.new(0, 0, -@c)
	end


	def sgn(num)
		return  0  if num.zero?
		return +1  if ( num > 0 )
		return -1
	end


	def mesh
		# create mesh
		numpts = @edges * @rings + 2       # number of points in the mesh
		numpoly = @edges * (@rings + 1)    # number of polygons in the mesh
		mesh = Geom::PolygonMesh.new(numpts, numpoly) 

		# Add pole points to mesh
		north_pole = mesh.add_point([0, 0, +@c])
		south_pole = mesh.add_point([0, 0, -@c])

		# Add the other points to the mesh
		v = 2.0 / @v_exp
		h = 2.0 / @h_exp
		index_array = []
		for i in 1..@rings do
			phi = i * Math::PI / (@rings + 1)   # polar angle
			cos_phi = sgn(Math.cos(phi)) * Math.cos(phi).abs**v
			sin_phi = sgn(Math.sin(phi)) * Math.sin(phi).abs**v
			indices = []
			for j in 0...@edges do
				theta = j * 2 * Math::PI / @edges       # azimuthal angle
				cos_theta = sgn(Math.cos(theta)) * Math.cos(theta).abs**h
				sin_theta = sgn(Math.sin(theta)) * Math.sin(theta).abs**h
				x = @a * cos_theta * sin_phi
				y = @b * sin_theta * sin_phi
				z = @c * cos_phi
				indices.push( mesh.add_point([x, y, z]) )
			end
			index_array.push indices
		end

		# Add top row of polygons
		i1 = index_array.first
		for j in 0...@edges do
			k = (j+1) % @edges
			mesh.add_polygon i1[j], i1[k], north_pole
		end

		# Add inbetween rows of polygons
		i1 = index_array[0]
		for i in 1...@rings do
			i2 = index_array[i]
			for j in 0...@edges do
				k = (j+1) % @edges
				# Since latatitude lines are parallel, these 4 points are co-planar.
				mesh.add_polygon i2[j], i2[k], i1[k], i1[j]
			end
			i1 = i2
		end

		# Add bottom row of polygons
		i1 = index_array.last
		for j in 0...@edges do
			k = (j+1) % @edges
			mesh.add_polygon south_pole, i1[k], i1[j]
		end

		return mesh
	end
	
end  # class Superellipsoid

end  # module RP


# Menu
#----------------------------------------------------------------------------#
unless file_loaded? File.basename(__FILE__) 
	add_separator_to_menu('Draw')
	UI.menu('Draw').add_item('Superellipsoid') { RP::Superellipsoid.dialog } 
end

file_loaded File.basename(__FILE__)